"""
moin2gitwiki context object - carries state between components

This contains the basic context object, which has various global
state information in it such as the logging objects.
"""
import logging.handlers
import sys
from pathlib import Path

import attr

from .users import Moin2GitUserSet


LOG_FILE = "moin2gitwiki.log"
FILE_FORMATTER = logging.Formatter(
    "%(asctime)s — %(name)s — %(levelname)s — %(funcName)s:%(lineno)d — %(message)s",
)
CONSOLE_FORMATTER = logging.Formatter(
    "%(asctime)s — %(name)s — %(levelname)s — %(funcName)s:%(lineno)d — %(message)s",
)
SYSLOG_FORMATTER = logging.Formatter("%(name)s: [%(levelname)s] %(message)s")


@attr.s(kw_only=True, slots=True)
class Moin2GitContext:
    """
    Moin2GitContext Context Object - holds state, logging, etc

    Called from the cli code.  Sets up all the common requirements.

    Attributes:
        debug:      if true we output more debugging chatter
        verbose:    if true we output more progress information
        syslog:     if true we additionally log to syslog at debug level
        logger:     Logging object
        moin_data:  Path of the MoinMoin data directory
        users:      Moin user set object

    """

    logger: logging.Logger = attr.ib()
    moin_data: Path = attr.ib()
    users: Moin2GitUserSet = attr.ib(default=None)
    syslog: bool = attr.ib(default=False)
    debug: bool = attr.ib(default=False)
    verbose: bool = attr.ib(default=False)

    @classmethod
    def create_context(cls, **kwargs):
        """
        Create the context object

        Builds the requirements for the context object and returns an object
        """
        if "logger" not in kwargs:
            logger = logging.getLogger("moin2gitwiki")
            kwargs["logger"] = logger
        if "user_map" in kwargs:
            user_map = kwargs["user_map"]
            del kwargs["user_map"]
        #
        # make the paths absolute
        kwargs["moin_data"] = Path(kwargs["moin_data"]).resolve(strict=True)
        #
        # build the context object
        context = cls(**kwargs)
        context.configure_logger()
        #
        # get the users
        if user_map is not None:
            context.users = Moin2GitUserSet.load_users_from_file(
                path=user_map,
                logger=context.logger,
            )
        else:
            context.users = Moin2GitUserSet.load_users_from_wiki_data(
                wiki_data_path=context.moin_data,
                logger=context.logger,
            )
        #
        return context

    def get_file_handler(self) -> logging.handlers.TimedRotatingFileHandler:
        """
        Sets up and returns the file logging handler

        Returns:
            file_handler: logger file handler
        """
        file_handler = logging.handlers.TimedRotatingFileHandler(
            LOG_FILE,
            when="midnight",
        )
        file_handler.setLevel(logging.DEBUG)
        file_handler.setFormatter(FILE_FORMATTER)
        return file_handler

    def configure_logger(self):
        logger = self.logger
        logger.setLevel(logging.DEBUG)
        #
        # set up the console logging
        console_handler = logging.StreamHandler(sys.stdout)
        if self.debug:
            console_handler.setLevel(logging.DEBUG)
        elif self.verbose:
            console_handler.setLevel(logging.INFO)
        else:
            console_handler.setLevel(logging.WARNING)
        console_handler.setFormatter(CONSOLE_FORMATTER)
        logger.addHandler(console_handler)
        #
        # set up syslog
        if self.syslog:
            syslog_handler = logging.handlers.SysLogHandler(
                address="/dev/log",
                facility=logging.handlers.SysLogHandler.LOG_LOCAL0,
            )
            syslog_handler.setLevel(logging.DEBUG)
            syslog_handler.setFormatter(SYSLOG_FORMATTER)
            logger.addHandler(syslog_handler)
        logger.addHandler(self.get_file_handler())


# end



import json
import os
import re

import attr


@attr.s(kw_only=True, frozen=True, slots=True)
class Moin2GitUser:
    """
    Represents a Moin user - for mapping to git user commits

    Attributes:
        moin_id: MoinMoin user id - multi-component numeric string
        moin_name: The Moin username  - used as a git name
        email: Email address of the user account

    """

    moin_id: str = attr.ib()
    moin_name: str = attr.ib()
    email: str = attr.ib(default="user@example.org")

    @classmethod
    def load_user_from_file(cls, path, logger):
        """
        Reads the user data in from a moin user config file
        """
        with open(path) as f:
            data = f.read()
        moin_id = os.path.basename(path)
        user_dict = dict(re.findall(r"^([a-z_]+)=(.*)$", data, flags=re.MULTILINE))
        params = {"moin_id": moin_id, "moin_name": user_dict["name"]}
        if user_dict["email"] is not None and user_dict["email"] != "":
            params["email"] = re.sub("[^A-Za-z0-9@._-]", "", user_dict["email"])
        user = cls(**params)
        logger.debug(f"User added: {user.moin_name}")
        return user


@attr.s(kw_only=True, frozen=True, slots=True)
class Moin2GitUserSet:
    """
    Represents a set of Moin users for mapping into git

    Attributes:
        id_map: maps moin user ids to Moin2GitUser objects
        name_map: maps moin user names to Moin2GitUser objects
    """

    id_map: dict = attr.ib(default={})
    name_map: dict = attr.ib(default={})

    @classmethod
    def create_from_users(cls, users, logger):
        """
        Builds a Moin2GitUserSet from a list of Moin2GitUser objects
        """
        id_map = {}
        name_map = {}
        for user in users:
            id_map[user.moin_id] = user
            name_map[user.moin_name] = user
        # make sure we have an anonymous entry for things we cannot put a user to
        if "anonymous" not in name_map:
            anonymous = Moin2GitUser(
                moin_id="0000000000.00.00000",
                moin_name="anonymous",
                email="anonymous@example.org",
            )
            id_map[anonymous.moin_id] = anonymous
            name_map[anonymous.moin_name] = anonymous
        # package all the users into a set
        logger.debug("Building user set object")
        return cls(id_map=id_map, name_map=name_map)

    @classmethod
    def load_users_from_wiki_data(cls, wiki_data_path, logger):
        """
        Builds a Moin2GitUserSet from the wiki filesystem
        """
        users_dir = os.path.join(wiki_data_path, "user")
        logger.debug(f"Loading wiki users from {users_dir}")
        users = []
        for moin_id in os.listdir(users_dir):
            # check the moin id filename looks right
            if re.match(r"\d+\.\d+\.\d+$", moin_id):
                try:
                    logger.debug(f"Loading user id {moin_id}")
                    user = Moin2GitUser.load_user_from_file(
                        path=os.path.join(users_dir, moin_id),
                        logger=logger,
                    )
                    users.append(user)
                except OSError:
                    continue
        return cls.create_from_users(users=users, logger=logger)

    @classmethod
    def load_users_from_file(cls, path, logger):
        """
        Builds a Moin2GitUserSet from a saved json file
        """
        logger.debug(f"Loading wiki users from {path}")
        with open(path) as f:
            user_data_set = json.loads(f.read())
        users = []
        for entry in user_data_set:
            user = Moin2GitUser(**entry)
            users.append(user)
            logger.debug(f"Loaded user name {user.moin_name}")
        return cls.create_from_users(users=users, logger=logger)

    def save_users_to_file(self, path):
        """
        Writes a Moin2GitUserSet to a saved json file
        """
        user_data = []
        for user in self.name_map.values():
            user_data.append(attr.asdict(user))
        with open(path, "w") as f:
            json.dump(user_data, f, indent=2, sort_keys=True)

    def get_user_by_name(self, name):
        """
        Gets a Moin2GitUser by matching moin name
        """
        return self.name_map[name]

    def get_user_by_id(self, ident):
        """
        Gets a Moin2GitUser by matching a moin id
        """
        return self.id_map[ident]

    def get_user_by_id_or_anonymous(self, ident):
        """
        Gets a Moin2GitUser by matching a moin id.  If non-existant returns the anonymous id
        """
        if ident is None or ident not in self.id_map:
            return self.get_user_by_name("anonymous")
        else:
            return self.get_user_by_id(ident)


# end



import json
import uuid
from pathlib import Path

import attr
import requests


@attr.s(kw_only=True, slots=True)
class FetchCache:
    """
    Implements a local cache for URLs which can be persistant between runs

    Basic cache directory which contains an `index.json` file with a table of
    URLs and the cache file they map to.  Has zero intelligence - assumes
    everything can be cached for ever - which is reasonable considering the
    things we request via the cache.

    Attributes:
        cache_directory:    Path of the cache directory
        index_path:         Path of the cache index file - normally `index.json` within `cache_directory`
        cache_map:          The dict mapping URLs to filenames within the cache
        ctx:                Context object (used for logging etc)

    """

    cache_directory: Path = attr.ib()
    index_path: Path = attr.ib()
    cache_map: dict = attr.ib(default={})
    ctx = attr.ib(repr=False)

    @classmethod
    def initialise_cache(cls, cache_directory: Path, ctx):
        """
        Build and preload the cache object

        Creates if needed the passed `cache_directory`, and either loads the
        existing `index.json` or writes an empty one.
        """
        # ensure directory exists
        cache_directory.mkdir(mode=0o777, parents=True, exist_ok=True)
        # ensure we have it as an absolute path
        cache_directory = cache_directory.resolve(strict=True)
        #
        # load the index should it exist
        index_path = cache_directory.joinpath("index.json")
        cache_map = {}
        try:
            cache_map = json.loads(index_path.read_text())
        except (OSError, ValueError):
            # if anything goes wrong then we just ignore it and
            # write out a blank cache file
            cache_map = {}
            cls.write_index(index_path=index_path, cache_map=cache_map)
        #
        # build and return the object
        ctx.logger.debug(f"Building cache in directory {cache_directory}")
        return cls(
            cache_directory=cache_directory,
            index_path=index_path,
            cache_map=cache_map,
            ctx=ctx,
        )

    @classmethod
    def write_index(cls, index_path: Path, cache_map: dict):
        """Write the cache index out to disk"""
        index_path.write_text(json.dumps(cache_map, indent=2))

    def fetch(self, url: str) -> str:
        """Fetch a URL, from the cache if there, otherwise put a copy into cache"""
        #
        # is this in the cache already
        if url in self.cache_map:
            item_name = self.cache_map[url]
            item_path = self.cache_directory.joinpath(item_name)
            try:
                content = item_path.read_text()
                self.ctx.logger.debug(f"Retrieved {url} from cache")
                return content
            except OSError:
                pass  # just move on to refetch
        #
        # if you get here then the url is either not in the cache or we
        # failed to retrieve it off disk - in either case we just fetch it
        item_name = uuid.uuid4().hex
        item_path = item_path = self.cache_directory.joinpath(item_name)
        self.ctx.logger.debug(f"Fetching {url}")
        try:
            response = requests.get(url)
            content = response.text
        except OSError:
            self.ctx.logger.warning(f"No response to {url}")
            content = ""
        #
        # write to cache
        item_path.write_text(content)
        self.ctx.logger.debug(f"Wrote {url} to {item_name}")
        #
        # update cache index
        self.cache_map[url] = item_name
        self.write_index(index_path=self.index_path, cache_map=self.cache_map)
        #
        # return response content
        return content


# end



import typing

import attr

from .wikiindex import MoinEditEntry


@attr.s(kw_only=True, slots=True)
class GitExportStream:
    """
    Output a git fast-export formatted stream for each revision

    This object handles the state information to output the git commits for
    the Moin wiki revisions.

    Attributes:
        output:     The output file stream of git fast-export commands
        mark_number: The current git mark number
        last_commit_mark: The git mark number of the last commit
        ctx:        The context object - used for `logger` and `user` mapping

    """

    output: typing.BinaryIO = attr.ib()
    mark_number: int = attr.ib(default=1)
    last_commit_mark: int = attr.ib(default=None)
    branch: str = attr.ib(default="refs/heads/master")
    ctx = attr.ib(repr=False)

    def add_wiki_revision(
        self,
        revision: MoinEditEntry,
        content: bytes,
    ):
        """
        Add a wiki revision as a git commit

        Parameters:
            revision:   A wiki revision object
            content:    The content of the wiki object, after translation, as bytes

        """
        name = revision.markdown_page_path()
        if content is not None:
            blob_ref = self.output_blob(content)
        if self.last_commit_mark is None:
            self.write_string(f"reset {self.branch}\n")
        self.write_string(f"commit {self.branch}\n")
        commit_ref = self.write_next_mark()
        self.write_changer("author", revision)
        self.write_changer("committer", revision)
        if revision.comment != "":
            self.output_data_string(f"{revision.comment}\n")
        else:
            if content is not None:
                self.output_data_string(f"Add/Update {name}\n")
            else:
                self.output_data_string(f"Delete {name}\n")
        if self.last_commit_mark is not None:
            self.write_string(f"from :{self.last_commit_mark}\n")
        if content is None:
            self.write_string(f"D {name}\n\n")
        else:
            if (
                revision.previous_page_name is not None
                and revision.previous_page_name != revision.page_name
            ):
                self.write_string(
                    f"D {revision.markdown_transform(revision.previous_page_name)}\n",
                )
            self.write_string(f"M 100644 :{blob_ref} {name}\n\n")
        self.last_commit_mark = commit_ref
        self.ctx.logger.debug(f"Written commit {commit_ref}")

    def write_changer(self, what: str, revision: MoinEditEntry):
        """
        Add an author/committer entry with date

        Parameters:
            what:       Normally either `committer` or `author`
            revision:   A wiki revision object

        """
        self.write_string(
            f"{what} {revision.user.moin_name} <{revision.user.email}> {int(revision.edit_date.timestamp())} +0000\n",
        )

    def get_next_mark(self):
        """
        Increment and return the mark number
        """
        mark = self.mark_number
        self.mark_number += 1
        return mark

    def write_next_mark(self):
        """
        Write out the next mark number
        """
        mark = self.get_next_mark()
        self.write_string(f"mark :{mark}\n")
        return mark

    def output_blob(self, content: bytes):
        """
        Output a blob object

        Parameters:
            content:    The content of the blob, as bytes

        """
        self.output.write(b"blob\n")
        blob_ref = self.write_next_mark()
        self.output_data(content)
        return blob_ref

    def output_data(self, content: bytes):
        """
        Output a set of data bytes

        Parameters:
            content:    The content of data, as bytes

        """
        self.write_string(f"data {len(content)}\n")
        self.output.write(content)

    def write_string(self, string: str):
        """
        Write a string out with utf-8 encoding into bytes
        """
        self.output.write(string.encode("utf-8"))

    def output_data_string(self, string: str):
        """
        Write a string out as a data object with utf-8 encoding into bytes
        """
        self.output_data(string.encode("utf-8"))

    def end_stream(self):
        """
        Write the end of stream information
        """
        self.write_string(f"reset {self.branch}\n")
        self.write_string(f"from :{self.last_commit_mark}\n")


# end



"""
moin2gitwiki Command Line Handling

Documentation is in the commands part of the documentation - the general
internals handling does not parse click decorators very well :-(

"""
import os
import subprocess
from pathlib import Path

import click

from . import __version__
from .context import Moin2GitContext
from .gitrevision import GitExportStream
from .moin2markdown import Moin2Markdown
from .wikiindex import MoinEditEntries


# -----------------------------------------------------------------------
@click.group()
@click.option("--debug/--no-debug", default=False, envvar="MOIN2GIT_DEBUG")
@click.option("--verbose/--no-verbose", default=False, envvar="MOIN2GIT_VERBOSE")
@click.option("--syslog/--no-syslog", default=False, envvar="MOIN2GIT_SYSLOG")
@click.option(
    "--moin-data",
    type=click.Path(exists=True, file_okay=False, dir_okay=True),
    envvar="MOIN2GIT_DATA",
)
@click.option(
    "--user-map",
    type=click.Path(exists=True, file_okay=True, dir_okay=False),
    envvar="MOIN2GIT_USERS",
)
@click.version_option(__version__)
@click.pass_context
def moin2gitwiki(ctx, syslog, verbose, debug, moin_data, user_map):
    """
    MoinMoin To Git Wiki Tools Command Line Utility

    Converts a MoinMoin wiki into a git repository populated with Markdown
    formatted pages, set up for use on a git based wiki such as the built in
    wiki for `gitea`, `github` or `gitlab`

    This parses the users and the revision structure from the MoinMoin data
    filesystem.  However converting the wiki markup was found to be best done
    by converting the output HTML using `pandoc`.

    The utility requires `git` and `pandoc` commands to be available in the
    path.

    #### Environment Variables

    The common options can also be set by use of environment variables:

    - `--debug` - `MOIN2GIT_DEBUG` - Output debugging logging

    - `--verbose` - `MOIN2GIT_VERBOSE` - Output verbose logging

    - `--syslog` - `MOIN2GIT_SYSLOG` - Send logging to syslog

    - `--moin-data` - `MOIN2GIT_DATA` - Data directory for moin

    - `--user-map` - `MOIN2GIT_USERS` - User map for moin - see the `save-users` command for info

    - `cache-directory` - `MOIN2GIT_CACHE` - Directory for moin component fetches.
      This defaults to `_cache` in the current directory.

    #### Help

    Running the ``moin2gitwiki`` command on its own will show some help
    information.

    """
    ctx.obj = Moin2GitContext.create_context(
        syslog=syslog,
        debug=debug,
        verbose=verbose,
        moin_data=moin_data,
        user_map=user_map,
    )


# -----------------------------------------------------------------------
@moin2gitwiki.command()
@click.pass_obj
def check(ctx):
    """A minimal check to see if we run in this environment"""
    # we may improve on this...
    click.echo("System check identified no issues")
    ctx.logger.debug("check complete")


# -----------------------------------------------------------------------
@moin2gitwiki.command()
@click.argument("filename", type=click.Path(file_okay=True, dir_okay=False))
@click.pass_obj
def save_users(ctx, filename):
    """
    Write the user map out to a file

    This writes all the users found in the wiki out to a JSON file. This can
    then be modified, if required, and used as the input to the `--user-map`
    option - typically this would be to fix any email address or name issues.

    These user entries are used to set the author of git commits within the
    output repository.

    The file format is an JSON file consisting of an array of user records,
    which each look like:-

        {
            "email": "user@example.com",
            "moin_id": "1358271613.26.36417",
            "moin_name": "SomeUser"
        },

    """
    ctx.users.save_users_to_file(filename)


# -----------------------------------------------------------------------
@moin2gitwiki.command()
@click.option(
    "--cache-directory",
    default="_cache",
    envvar="MOIN2GIT_CACHE",
)
@click.option(
    "--url-prefix",
    "--prefix",
    default="http://localhost/jrtwiki/",
    envvar="MOIN2GIT_PREFIX",
)
@click.argument(
    "destination",
    type=click.Path(exists=False, file_okay=False, dir_okay=True),
)
@click.pass_obj
def fast_export(ctx, cache_directory, url_prefix, destination):
    """
    Git fast-export all the revisions in the wiki into markdown git wiki form

    Named for the `git fast-export` command, although it actually builds a new
    git repository and then translates each revision at a time into a command
    stream for `git-fast-import` on that new repository.  After all pages and
    revisions have been processed the new git wiki repo instance is garbage
    collected  (to compress all the revisions into a more compact set of git
    packs) and finally checked out.

    Page names are slightly modified - the "(2f)" seen in wiki file names
    (which is normally displayed as a `/` character) are changed to
    underscores.  Internal links are remapped - however if a link goes within
    the wiki namespace to something that was not found in the wiki (this may
    include attachments which are not currently bought across), then the link
    is deleted (although the link text is left).

    Although the filesystem data is read to derive the revision and history
    information, the actual page transformation is done by retrieving the
    page html from its webserver, cutting the content div out of that html,
    doing a few simplifications and translations (specifcially images
    corresponding to emojis are converted to emoji forms).  This HTML is then
    pass through pandoc to get a markdown (specifically github flavoured
    markdown).

    """
    # cwd = Path.cwd()
    destination = Path(destination)
    if destination.exists():
        raise SystemExit(f"Destination path {destination} already exists.")
    #
    # build your initial revision set from the wiki data
    revisions = MoinEditEntries.create_edit_entries(ctx=ctx)
    click.echo(click.style(f"Read {revisions.count()} wiki revisions", fg="green"))
    #
    # build a link_table
    link_table = {
        revision.page_name_unescaped(): revision.markdown_page_name()
        for revision in revisions.entries
    }
    #
    # build the translator
    translator = Moin2Markdown.create_translator(
        ctx=ctx,
        cache_directory=Path(cache_directory),
        url_prefix=url_prefix,
        link_table=link_table,
    )
    #
    # build the output git instance
    destination.mkdir(mode=0o755)
    os.chdir(destination)
    subprocess.run(["git", "init"])
    with subprocess.Popen(["git", "fast-import"], stdin=subprocess.PIPE) as gitstream:
        export = GitExportStream(output=gitstream.stdin, ctx=ctx)
        with click.progressbar(revisions.entries) as entries:
            for revision in entries:
                content = translator.retrieve_and_translate(revision=revision)
                export.add_wiki_revision(
                    revision=revision,
                    content=content,
                )
            revision, content = revisions.create_home_page()
            export.add_wiki_revision(revision=revision, content=content.encode("utf-8"))
        export.end_stream()
    subprocess.run(["git", "gc", "--aggressive"])  # pack it
    subprocess.run(["git", "checkout", "master"])  # check out the data


# -----------------------------------------------------------------------
@moin2gitwiki.command()
@click.option(
    "--cache-directory",
    default="_cache",
    envvar="MOIN2GIT_CACHE",
)
@click.option(
    "--url-prefix",
    "--prefix",
    default="http://localhost/jrtwiki/",
    envvar="MOIN2GIT_PREFIX",
)
@click.argument("page", required=True, type=str)
@click.argument("version", required=True, type=int)
@click.pass_obj
def translate_page(ctx, cache_directory, url_prefix, page, version):
    """
    Fetch a single page revision and translate to Markdown

    The first argument is a page name, the second an integer revision.

    The translation process is as described for the `fast-expor`a command.
    """
    #
    # build your initial revision set from the wiki data
    revisions = MoinEditEntries.create_edit_entries(ctx=ctx)
    click.echo(click.style(f"Read {revisions.count()} wiki revisions", fg="green"))
    #
    # build a link_table
    link_table = {
        revision.page_name_unescaped(): revision.markdown_page_name()
        for revision in revisions.entries
    }
    #
    # build the translator
    translator = Moin2Markdown.create_translator(
        ctx=ctx,
        cache_directory=Path(cache_directory),
        url_prefix=url_prefix,
        link_table=link_table,
    )
    #
    # find the page and translate it
    for revision in revisions.entries:
        if revision.page_name == page and int(revision.page_revision) == version:
            content = translator.retrieve_and_translate(revision=revision)
            print(content)


# -----------------------------------------------------------------------
# end



import os
import re
from datetime import datetime
from datetime import timedelta
from typing import Tuple

import attr

from .users import Moin2GitUser


@attr.s(kw_only=True, frozen=True, slots=True)
class MoinEditEntry:
    """
    Represents a Moin page revision

    There are multiple revisions per page.

    Attributes:
        edit_date: The date of the edit
        page_revision: The revision id of this revision - a string of a zero padded number
        edit_type: Moin edit type
        page_name: The name of the page from the index file
        previous_page_name: The name the page previously had if renamed
        page_path: The name on the filesystem of the page
        attachment: attachment field - not used
        comment: comment filed - only used for git comments
        user: the mapped moin user
        ctx: Context - there for moin_path and logging

    """

    edit_date: datetime = attr.ib()
    page_revision: str = attr.ib()
    edit_type: str = attr.ib()
    page_name: str = attr.ib()
    previous_page_name: str = attr.ib(default=None)
    page_path: str = attr.ib()
    attachment: str = attr.ib(default="")
    comment: str = attr.ib(default="")
    user: Moin2GitUser = attr.ib()
    ctx = attr.ib(repr=False)

    def wiki_content_path(self):
        """The file pathname of the revision file"""
        return self.ctx.moin_data.joinpath(
            "pages",
            self.page_path,
            "revisions",
            self.page_revision,
        )

    def wiki_content_bytes(self):
        """The content of the wiki revision retrieved as a byte string"""
        lines = self.wiki_content()
        if lines is None:
            return lines
        else:
            lines.append("")
            return "\n".join(lines).encode("utf-8")

    def wiki_content(self):
        """The content of the wiki revision as an array of strings"""
        lines = []
        try:
            lines = self.wiki_content_path().read_text().splitlines(keepends=False)
        except OSError:
            lines = None
        return lines

    def unescape(self, thing: str) -> str:
        """Uescape a wiki name - translate (2f) to /"""
        return thing.replace("(2f)", "/")

    def page_name_unescaped(self) -> str:
        """Unescape the page name"""
        return self.unescape(self.page_name)

    def page_path_unescaped(self) -> str:
        """Unescape the page path"""
        return self.unescape(self.page_path)

    def markdown_transform(self, thing: str) -> str:
        """Translates the (2f) to _ for use in Markdown page names"""
        return thing.replace("(2f)", "_")

    def markdown_page_path(self):
        """Page path translated"""
        return self.markdown_transform(self.page_name) + ".md"

    def markdown_page_name(self):
        """Page name translated"""
        return self.markdown_transform(self.page_name)


@attr.s(kw_only=True, frozen=True, slots=True)
class MoinEditEntries:
    """
    A sorted collection of Moin revision entry objects
    """

    entries: list = attr.ib()
    ctx = attr.ib(repr=False)

    @classmethod
    def create_edit_entries(cls, ctx):
        pages_dir = os.path.join(ctx.moin_data, "pages")
        pages = os.listdir(pages_dir)
        epoch = datetime(1970, 1, 1)
        entries = []
        for page in pages:
            ctx.logger.debug(f"Reading page {page}")
            edit_log_file = os.path.join(pages_dir, page, "edit-log")
            # read the edit-log file
            try:
                with open(edit_log_file) as f:
                    edit_log_data = f.readlines()
            except OSError:
                ctx.logger.warning(f"No edit-log for page {page}")
                continue
            # read the lines in the edit-log file
            previous_page_name = None
            for edit_line in edit_log_data:
                if not re.match(r"\d{15}", edit_line):  # check its an edit entry
                    continue
                # extract the fields out the edit entry
                edit_fields = edit_line.rstrip("\n").split("\t")
                edit_date = epoch + timedelta(microseconds=int(edit_fields[0]))
                edit_type = edit_fields[2]
                if edit_type in ("SAVENEW", "SAVE"):
                    entry = MoinEditEntry(
                        edit_date=edit_date,
                        page_revision=edit_fields[1],
                        edit_type=edit_type,
                        page_name=edit_fields[3],
                        previous_page_name=previous_page_name,
                        attachment=edit_fields[7],
                        comment=edit_fields[8],
                        page_path=page,
                        user=ctx.users.get_user_by_id_or_anonymous(edit_fields[6]),
                        ctx=ctx,
                    )
                    entries.append(entry)
        ctx.logger.debug("Sorting edit entries")
        entries.sort(key=lambda x: x.edit_date)
        ctx.logger.debug("Building edit entries object")
        return cls(
            entries=entries,
            ctx=ctx,
        )

    def count(self) -> int:
        return len(self.entries)

    def create_home_page(self) -> Tuple[MoinEditEntry, str]:
        """Builds a synthetic home page to link all the wiki entries together"""
        revision = MoinEditEntry(
            edit_date=datetime.now(),
            page_revision="1",
            edit_type="SAVENEW",
            page_name="Home",
            attachment="",
            comment="Synthetic Home Page",
            page_path="Home",
            user=self.ctx.users.get_user_by_id_or_anonymous("0"),
            ctx=self.ctx,
        )
        pages = set()
        for entry in self.entries:
            if entry.page_name.find("(2f)") < 0:
                pages.add(entry.markdown_page_name())
        content = "# Home Page\n\n"
        for item in sorted(pages):
            content += f"- [{item}]({item})\n"
        content += "\n----\n"

        return (revision, content)


# end



import re
import subprocess
from pathlib import Path
from typing import Optional

import attr
from bs4 import BeautifulSoup
from furl import furl

from .fetch_cache import FetchCache
from .wikiindex import MoinEditEntry


def is_a_linemark_para(tag):
    return (
        tag.name == "p"
        and tag.has_attr("class")
        and re.match(r"line\\d+", tag["class"][0])
    )


@attr.s(kw_only=True, frozen=True, slots=True)
class Moin2Markdown:
    """
    Conversion object to convert MoinMoin wiki markup to Markdown

    Attributes:
        fetch_cache:    A FetchCache object used to retrieve URLs
        url_prefix:     The URL prefix of the Moin wiki web presence
        link_table:     A mapping of Moin unescaped names to page names
        ctx:            Context object - logger and user mapping etc
    """

    #
    # -- attributes
    fetch_cache: FetchCache = attr.ib()
    url_prefix: furl = attr.ib()
    link_table: dict = attr.ib()
    ctx = attr.ib(repr=False)
    #
    # smiley mapping
    smiley_map = {
        "X-(": ":rage:",
        ":(": ":confused:",
        ";)": ":wink:",
        ":-?": ":stuck_out_tongue:",
        ":-(": ":frowning_face:",
        ";-)": ":wink:",
        "{X}": ":x:",
        "{3}": ":three:",
        ":D": ":grin:",
        ":)": ":slightly_smiling_face:",
        "/!\\": ":warning:",
        ":\\": ":confounded:",
        ":-)": ":smiley:",
        "|-)": ":pensive:",
        "{i}": ":information_source:",
        "{*}": ":star:",
        "<:(": ":mask:",
        "B)": ":sunglasses:",
        "<!>": ":warning:",
        ">:>": ":imp:",
        "B-)": ":nerd_face:",
        "(./)": ":white_check_mark:",
        "{1}": ":one:",
        "{o}": ":star:",
        ":o": ":anguished:",
        ":))": ":rofl:",
        "(!)": ":bulb:",
        "|)": ":monocle_face:",
        ":-))": ":rofl:",
        "{OK}": ":ok:",
        "{2}": ":two:",
    }

    @classmethod
    def create_translator(
        cls,
        ctx,
        cache_directory: Path,
        url_prefix: str,
        link_table: dict,
    ):
        """
        Build a translator object

        Parameters:
            ctx:            Context object (logger etc)
            cache_directory:    Path object for the cache directory
            url_prefix:     The base URL for the MoinMoin wiki
            link_table:     A translation table for wiki links

        """
        #
        # Build a fetch cache
        fetch_cache = FetchCache.initialise_cache(
            cache_directory=cache_directory,
            ctx=ctx,
        )
        return cls(
            fetch_cache=fetch_cache,
            link_table=link_table,
            url_prefix=furl(url_prefix),
            ctx=ctx,
        )

    def retrieve_and_translate(self, revision: MoinEditEntry) -> Optional[str]:
        """
        Retrieve a wiki revision, and translate it to markdown

        Parameters:
            revision:    The wiki revision object for the revision we want

        If the revision maps to an empty object - ie it deleted the page, or
        similar, then a None object is returned.

        """
        # check if this revision has any content...
        lines = revision.wiki_content()
        if lines is None:
            return None
        else:
            target = self.url_prefix.copy()
            target /= revision.page_path_unescaped()
            target.args["action"] = "recall"
            target.args["rev"] = revision.page_revision
            content = self.fetch_cache.fetch(target.url)
            main_content = self.extract_content_section(content)
            translated = self.translate(main_content)
            return translated

    def extract_content_section(self, html: str) -> str:
        """
        Extract the content part of the HTML, and simplify

        Parameters:
            html:    The html data

        Pulls out the content div and simplifies the  HTML.
        Simplification consists of:-

        - stripping out redundant anchor spans
        - remove the additional line marking paragraphs
        - rewrite a/hrefs
        - strip internal a/hrefs that have no existng target
        - strip class attributes from links
        - remap any emoji img to the emoji sequence

        """
        soup = BeautifulSoup(html, "html.parser")
        content = soup.find(id="content")
        #
        # now strip out excess rubbish - anchor spans
        for tag in content.find_all(class_="anchor"):
            tag.decompose()
        #
        # Remove dead <p class="line???"> with no closer
        for tag in content.find_all(is_a_linemark_para):
            tag.unwrap()
        #
        # now find all the links, and if within the wiki, rewrite
        for tag in content.find_all("a"):
            target = tag["href"]
            if target:
                url = self.url_prefix.copy().join(target)
                if url.url.startswith(self.url_prefix.url):
                    new_url = url.url[len(self.url_prefix.url) :]
                    if new_url in self.link_table:
                        tag["href"] = self.link_table[new_url]
                    else:
                        tag.unwrap()
            #
            # strip any class attributes on links - tend to upset the translator
            if tag.has_attr("class"):
                del tag["class"]
        #
        # This might not always work but removing all <div>s makes output cleaner
        for tag in content.find_all("div"):
            tag.unwrap()
        #
        # now find all the images and see if they map to emojis
        # MoinMoin puts the emoji code in the title, so will purely match on that
        for tag in content.find_all("img"):
            if tag.has_attr("title") and tag["title"] in self.smiley_map:
                tag.replace_with(" " + self.smiley_map[tag["title"]] + " ")

        return "".join([str(x) for x in content.contents])

    def translate(self, input: str) -> str:
        """Translate HTML to Github Flavoured Markdown using pandoc"""
        process = subprocess.Popen(
            ["pandoc", "-f", "html", "-t", "gfm"],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
        )
        (output, _) = process.communicate(input.encode("utf-8"))
        return output.decode("utf-8")


# end



#!/usr/bin/env python
"""Tests for `moin2gitwiki` package."""
import re

import pytest
from click.testing import CliRunner

from moin2gitwiki import __version__
from moin2gitwiki import cli


@pytest.fixture
def response():
    """Sample pytest fixture.

    See more at: http://doc.pytest.org/en/latest/fixture.html
    """
    # import requests
    # return requests.get("https://github.com/audreyr/cookiecutter-pypackage")


def test_content(response):
    """Sample pytest test function with the pytest fixture as an argument."""
    # from bs4 import BeautifulSoup
    # assert "GitHub" in BeautifulSoup(response.content).title.string


def test_command_line_interface():
    """Test the CLI."""
    runner = CliRunner()
    result = runner.invoke(cli.moin2gitwiki)
    assert result.exit_code == 0
    assert "moin2gitwiki" in result.output
    help_result = runner.invoke(cli.moin2gitwiki, ["--help"])
    assert help_result.exit_code == 0
    assert re.search(r"--help \s+ Show this message and exit.", help_result.output)
    version_result = runner.invoke(cli.moin2gitwiki, ["--version"])
    assert version_result.exit_code == 0
    assert f", version {__version__}" in version_result.output



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['moin2gitwiki']

package_data = \
{'': ['*']}

install_requires = \
['attrs>=20.3.0,<21.0.0',
 'beautifulsoup4>=4.9.3,<5.0.0',
 'click>=7.1.2,<8.0.0',
 'furl>=2.1.0,<3.0.0',
 'requests[socks]>=2.25.1,<3.0.0']

entry_points = \
{'console_scripts': ['moin2gitwiki = moin2gitwiki.cli:moin2gitwiki']}

setup_kwargs = {
    'name': 'moin2gitwiki',
    'version': '0.3.0',
    'description': 'MoinMoin To Git (Markdown) Wiki Converter',
    'long_description': '# MoinMoin To Git (Markdown) Wiki Converter\n\n[![ci](https://img.shields.io/travis/nigelm/moin2gitwiki.svg)](https://travis-ci.com/nigelm/moin2gitwiki)\n[![documentation](https://img.shields.io/badge/docs-mkdocs%20material-blue.svg?style=flat)](https://nigelm.github.io/moin2gitwiki/)\n[![pypi version](https://img.shields.io/pypi/v/moin2gitwiki.svg)](https://pypi.python.org/pypi/moin2gitwiki)\n\nApp to convert a MoinMoin wiki file tree into a git based wiki as used on\ngithub, gitlab or gitea.\n\n## Current Version\n\nVersion: `0.3.0`\n\n## Translation Method\n\nOriginally the intention was to translate purely by converting the MoinMoin\nmarkup to markdown markup - using the MoinMoin data retrieved from the\nfilesystem.\n\nHowever, although it makes determining the overall page list and revision list\nmuch easier, it was found that translating the wiki markup at this level was\ntoo complex and fragile for this to work without a huge amount of special\ncasing.\n\nSo, after the revision structure is derived from the filesystem, each page\nrevision is retrieved by http requests to the running MoinMoin wiki.  This is\nthen reduced to just the page content (by picking out the content div from the\nhtml), and some light editing applied to simplify the HTML - specifically:-\n\n- Remove the anchor spans that MoinMoin adds - these add no visual or\n  readable content, but confuse the translator\n- Remove paragraph entries with CSS classes that start `line` - these\n  again appear to be for non-required purposes (likely for showing diffs\n  between revisions) - and they break the translator\n- Fix links that point within the wiki - if the target does not exist\n  then the text is left but the link removed.\n- Strips CSS classes off links - again these upset the translator\n- Translate any images that appear to be MoinMoin emoji characters (which\n  are rendered as images) into gollum emoji characters\n\nThis simplified HTML is then passed through the pandoc command:-\n\n    pandoc -f html -t gfm\n\nAnd the resulting Github flavoured Markdown is taken as the new form.\n\nThis handles the vast majority of normal markup correctly, including lists and\nmany types of tables.  Some complicated markup or complex tables end up being\npassed through as HTML - which displays correctly but is less easy to parse\nand edit.\n\n## Issues\n\nThe overall process is not particularly fast.  But this should be something\nyou only do once (or a few attempts) so raw speed is not needed.\n\n## Installation\n\nI have now made this available as a pypi package, in which case it can be\ninstalled by running\n\n    pip install moin2gitwiki\n\nHowever to use it you will also need to install the `pandoc` and `git`\npackages as these commands are run during the conversion.\n\nHowever it can be installed from the repo - it uses\n[`poetry`](https://python-poetry.org/) to manage dependancies etc, so the best\nway to make use of this is to install [`poetry`](https://python-poetry.org/)\nfor your python version and then:-\n\n    poetry install\n\nthe command can then be run as\n\n    poetry run moin2gitwiki ...\n\n## Todo\n\n- Make tests effective\n\n----\n',
    'author': 'Nigel Metheringham',
    'author_email': 'nigelm@cpan.org',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/nigelm/moin2gitwiki',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
